#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "bitmanip.h"
#include "twiddles.h"
#include "fft_r4_spfp.h"
#include "taps.h"

static void radix_4_butterfly(
	complex_float   & A,
	complex_float   & B,
	complex_float   & C,
	complex_float   & D,
	complex_float     u,
	complex_float     v,
	fft_opts_t      & opts,
	fft_verbosity_t & vopts);

// ----------------------------------------------------------------
void fft_r4_spfp(
	complex_float * y,
	fft_opts_t      opts,
	fft_verbosity_t vopts)
{
	int N = opts.N;
	int log2N = splog2((unsigned)N);

	if (!is_a_power_of_four(N)) {
		std::cerr << "fft_r4_spfp: input length "
			<< N << " is not a power of four.\n";
		exit(1);
	}

	tvector<complex_float> W = half_twiddle_table_float(N, opts.forward);
	int i;
	int num_stages, stageno;
	int num_flocks, fno;
	int flock_sep, wingspan;
	int bperflock, bno;
	int left_twiddle_stride, right_twiddle_stride;
	int A, B, C, D, u, v;

	num_stages = log2N >> 1;

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	if (vopts.do_taps)
		complex_float_tap(y, opts.N,
			"host_fft_r4_in.tap");

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Output folding.
	if (opts.fold_out) {
		for (i = 1; i < N; i+=2)
			y[i] = -y[i];
	}

	if (vopts.do_taps)
		complex_float_tap(y, opts.N,
			"host_fft_r4_post_fold_out.tap");

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Bit-reverse stage.
	for (i = 0; i < N; i++) {
		int ir = bit_reverse(i, log2N);
		complex_float temp;
		if (i < ir) {
			temp  = y[i];
			y[i]  = y[ir];
			y[ir] = temp;
		}
	}

	if (vopts.do_taps)
		complex_float_tap(y, opts.N,
			"host_fft_r4_post_bit_rev.tap");

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Radix-four stages.

	num_flocks = N/4;
	flock_sep  = 4;
	bperflock  = 1;
	wingspan   = 1;
	left_twiddle_stride = N/2;
	right_twiddle_stride = N/4;

	for (stageno = 0; stageno < num_stages; stageno++) {

		if (vopts.do_taps)
			complex_float_tap(y, opts.N,
				"host_fft_r4_stage_%d_in.tap", stageno);

		if (vopts.verbose) {
			printf("\n");
			printf("Host r4 spfp stage %d\n", stageno);
		}

		for (fno = 0; fno < num_flocks; fno++) {
			u = 0;
			v = 0;

			for (bno = 0; bno < bperflock; bno++) {
				A = fno * flock_sep + bno;
				B = A + wingspan;
				C = B + wingspan;
				D = C + wingspan;

				if (vopts.verbose) {
					printf("H %3d %3d %3d %3d  %3d %3d\n",
						A, B, C, D, u, v);
				}

				if (vopts.extra_verbose) {
					printf(" HT u=%11.7f,%11.7f v=%11.7f,%11.7f\n",
						W[u].re, W[u].im, W[v].re, W[v].im);
					printf(" HI A=%11.7f,%11.7f B=%11.7f,%11.7f C=%11.7f,%11.7f D=%11.7f,%11.7f\n",
						y[A].re, y[A].im, y[B].re, y[B].im,
						y[C].re, y[C].im, y[D].re, y[D].im);
				}

				radix_4_butterfly(y[A], y[B], y[C], y[D],
					W[u], W[v], opts, vopts);

				if (vopts.extra_verbose) {
					printf(" HO A=%11.7f,%11.7f B=%11.7f,%11.7f C=%11.7f,%11.7f D=%11.7f,%11.7f\n",
						y[A].re, y[A].im, y[B].re, y[B].im,
						y[C].re, y[C].im, y[D].re, y[D].im);
				}

				if (opts.scale) {
					y[A] *= 0.5;
					y[B] *= 0.5;
					y[C] *= 0.5;
					y[D] *= 0.5;
				}

				u += left_twiddle_stride;
				v += right_twiddle_stride;
			}
		}

		num_flocks >>= 2;
		flock_sep  <<= 2;
		bperflock  <<= 2;
		wingspan   <<= 2;
		left_twiddle_stride >>= 2;
		right_twiddle_stride >>= 2;

		if (vopts.do_taps)
			complex_float_tap(y, opts.N,
				"host_fft_r4_stage_%d_out.tap", stageno);
	}

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Input folding.

	if (vopts.do_taps)
		complex_float_tap(y, opts.N,
			"host_fft_r4_pre_fold_in.tap");

	if (opts.fold_in) {
		for (i = 1; i < N; i+=2)
			y[i] = -y[i];
	}

	if (vopts.do_taps)
		complex_float_tap(y, opts.N,
			"host_fft_r4_out.tap");
}

// ----------------------------------------------------------------
static void radix_4_butterfly(
	complex_float   & A,
	complex_float   & B,
	complex_float   & C,
	complex_float   & D,
	complex_float     u,
	complex_float     v,
	fft_opts_t      & opts,
	fft_verbosity_t & vopts)
{
	complex_float uB   = u * B;
	complex_float uD   = u * D;
	complex_float ApuB = A + uB;
	complex_float AmuB = A - uB;
	complex_float CpuD = C + uD;
	complex_float CmuD = C - uD;

	complex_float pmi;
	if (opts.forward)
		pmi = complex_float(0.0, -1.0);
	else
		pmi = complex_float(0.0,  1.0);
	complex_float    vCpuD = v * CpuD;
	complex_float pmivCmuD = pmi * v * CmuD;

	if (vopts.butterfly_verbose) {
		printf(" HBF u=%11.7f,%11.7f v=%11.7f,%11.7f\n",
			u.re, u.im, v.re, v.im);
		printf(" HBF A=%11.7f,%11.7f B=%11.7f,%11.7f C=%11.7f,%11.7f D=%11.7f,%11.7f\n",
			A.re, A.im, B.re, B.im, C.re, C.im, D.re, D.im);
	}

	A = ApuB +    vCpuD;
	B = AmuB + pmivCmuD;
	C = ApuB -    vCpuD;
	D = AmuB - pmivCmuD;

	if (vopts.butterfly_verbose) {
		printf(" HBF A=%11.7f,%11.7f B=%11.7f,%11.7f C=%11.7f,%11.7f D=%11.7f,%11.7f\n",
			A.re, A.im,
			B.re, B.im,
			C.re, C.im,
			D.re, D.im);
		printf("\n");
	}

}
