#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "complex_float.h"
#include "complex_double.h"
#include "gasdev.h"
#include "tmatrix.h"

typedef struct _opts_t {
	int    do_seed;
	int    seed;
	int    do_complex;
	int    do_double;
	float  mean;
	float  stddev;
	int    num_rows;
	int    num_cols;
	char * output_file_name;
} opts_t;

static void set_defaults(opts_t * popts);
static int parse_command_line(int argc, char ** argv, opts_t * popts);

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr,
		"Usage: %s [options] {# rows} {# cols} {output file}\n",
		argv0);
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "-r for real, -c for complex.\n");
	fprintf(stderr, "-sp for single precision, -dp for double precision.\n");
	fprintf(stderr, "-s {seed}\n");
	fprintf(stderr, "-m {mean}\n");
	fprintf(stderr, "-t {stddev}\n");
	fprintf(stderr, "Seed is optional and defaults to time of day.\n");
	fprintf(stderr, "Mean is optional and defaults to 0.0.\n");
	fprintf(stderr, "Stddev is optional and defaults to 1.0.\n");
	exit(1);
}

// ----------------------------------------------------------------
int main(
	int     argc,
	char ** argv)
{
	opts_t opts;
	int i, j;

	set_defaults(&opts);
	if (!parse_command_line(argc, argv, &opts))
		usage(argv[0]);

	if (opts.do_seed)
		sran32(opts.seed);

	if (opts.do_double) {
		if (opts.do_complex) {
			tmatrix<complex_double> rand(opts.num_rows, opts.num_cols);
			for (i = 0; i < opts.num_rows; i++)
				for (j = 0; j < opts.num_cols; j++)
					rand[i][j] = complex_double_gasdev_ms(opts.mean, opts.stddev);
			rand.store_to_binary_file(opts.output_file_name);
		}
		else {
			tmatrix<double> rand(opts.num_rows, opts.num_cols);
			for (i = 0; i < opts.num_rows; i++)
				for (j = 0; j < opts.num_cols; j++)
					rand[i][j] = double_gasdev_ms(opts.mean, opts.stddev);
			rand.store_to_binary_file(opts.output_file_name);
		}
	}
	else {
		if (opts.do_complex) {
			tmatrix<complex_float> rand(opts.num_rows, opts.num_cols);
			for (i = 0; i < opts.num_rows; i++)
				for (j = 0; j < opts.num_cols; j++)
					rand[i][j] = complex_float_gasdev_ms(opts.mean, opts.stddev);
			rand.store_to_binary_file(opts.output_file_name);
		}
		else {
			tmatrix<float> rand(opts.num_rows, opts.num_cols);
			for (i = 0; i < opts.num_rows; i++)
				for (j = 0; j < opts.num_cols; j++)
					rand[i][j] = float_gasdev_ms(opts.mean, opts.stddev);
			rand.store_to_binary_file(opts.output_file_name);
		}
	}

	return 0;
}

// ----------------------------------------------------------------
static void set_defaults(opts_t * popts)
{
	popts->do_seed     = 0;
	popts->seed        = 0;
	popts->do_complex  = 1;
	popts->do_double   = 0;
	popts->mean        = 0.0;
	popts->stddev      = 1.0;
	popts->num_rows    = 1;
	popts->num_cols    = 1;
	popts->output_file_name = 0;
}

// ----------------------------------------------------------------
static int parse_command_line(int argc, char ** argv, opts_t * popts)
{
	int argi;
	for (argi = 1; argi < argc; argi++) {
		if (strcmp(argv[argi], "--help") == 0)
			return 0;
		if (argv[argi][0] != '-')
			break;

		if (strcmp(argv[argi], "-s") == 0) {
			argi++;
			if ((argc - argi) < 1)
				return 0;
			if (sscanf(argv[argi], "%i", &popts->seed) != 1)
				return 0;
			popts->do_seed = 1;
		}

		else if (strcmp(argv[argi], "-m") == 0) {
			argi++;
			if ((argc - argi) < 1)
				return 0;
			if (sscanf(argv[argi], "%f", &popts->mean) != 1)
				return 0;
		}

		else if (strcmp(argv[argi], "-t") == 0) {
			argi++;
			if ((argc - argi) < 1)
				return 0;
			if (sscanf(argv[argi], "%f", &popts->stddev) != 1)
				return 0;
		}

		else if (strcmp(argv[argi], "-r") == 0) {
			popts->do_complex = 0;
		}
		else if (strcmp(argv[argi], "-c") == 0) {
			popts->do_complex = 1;
		}

		else if (strcmp(argv[argi], "-dp") == 0) {
			popts->do_double = 1;
		}
		else if (strcmp(argv[argi], "-sp") == 0) {
			popts->do_double = 0;
		}

		else if (strcmp(argv[argi], "-sr") == 0) {
			popts->do_double  = 0;
			popts->do_complex = 0;
		}
		else if (strcmp(argv[argi], "-dr") == 0) {
			popts->do_double  = 1;
			popts->do_complex = 0;
		}
		else if (strcmp(argv[argi], "-sc") == 0) {
			popts->do_double  = 0;
			popts->do_complex = 1;
		}
		else if (strcmp(argv[argi], "-dc") == 0) {
			popts->do_double  = 1;
			popts->do_complex = 1;
		}

		else {
			return 0;
		}
	}
	if ((argc - argi) != 3)
		return 0;
	if (sscanf(argv[argi], "%d", &popts->num_rows) != 1)
		return 0;
	if (sscanf(argv[argi+1], "%d", &popts->num_cols) != 1)
		return 0;
	popts->output_file_name = argv[argi+2];

	return 1;
}
