#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "complex_float.h"
#include "tmatrix.h"

typedef struct _opts_t {
	int    do_complex;
	int    do_circular;
	float  num_row_bumps;
	float  num_col_bumps;
	int    num_rows;
	int    num_cols;
	char * output_file_name;
} opts_t;

static void set_defaults(opts_t * popts);
static int parse_command_line(int argc, char ** argv, opts_t * popts);

// ----------------------------------------------------------------
static float sinc(float x)
{
	if (x == 0.0)
		return 1.0;
	else
		return sin(x) / x;
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s [-circ] [-r|-c] {# row umps} {# col bumps} {# rows} {# cols} "
		"{output file}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(
	int     argc,
	char ** argv)
{
	opts_t opts;
	int i, j;

	set_defaults(&opts);
	if (!parse_command_line(argc, argv, &opts))
		usage(argv[0]);

	tmatrix<float> fdata(opts.num_rows, opts.num_cols);

	for (i = 0; i < opts.num_rows; i++) {
		// Scale from the range [0 .. num_rows) to the range
		// [-num_row_bumps .. num_row_bumps]
		float ifrac  = (float)i / (float)opts.num_rows;
		float iscale = -opts.num_row_bumps + 2.0 * ifrac * opts.num_row_bumps;


		for (j = 0; j < opts.num_cols; j++) {
			// Scale from the range [0 .. num_cols) to the range
			// [-num_col_bumps .. num_col_bumps]
			float jfrac  = (float)j / (float)opts.num_cols;
			float jscale = -opts.num_col_bumps + 2.0 * jfrac * opts.num_col_bumps;

			if (opts.do_circular) {
				float r = sqrt(iscale*iscale + jscale*jscale);
				fdata[i][j] = sinc(M_PI * r);
			}
			else {
				fdata[i][j] = sinc(M_PI * iscale) * sinc(M_PI * jscale);
			}
		}
	}

	if (opts.do_complex) {
		tmatrix<complex_float> cfdata(opts.num_rows, opts.num_cols);
		for (i = 0; i < opts.num_rows; i++)
			for (j = 0; j < opts.num_cols; j++)
				cfdata[i][j] = fdata[i][j];
		cfdata.store_to_binary_file(opts.output_file_name);
	}
	else {
		fdata.store_to_binary_file(opts.output_file_name);
	}

	return 0;
}

// ----------------------------------------------------------------
static void set_defaults(opts_t * popts)
{
	popts->do_complex       = 1;
	popts->do_circular      = 0;
	popts->num_row_bumps    = 1.0;
	popts->num_col_bumps    = 1.0;
	popts->num_rows         = 1;
	popts->num_cols         = 1;
	popts->output_file_name = 0;
}

// ----------------------------------------------------------------
static int parse_command_line(int argc, char ** argv, opts_t * popts)
{
	int argi;
	for (argi = 1; argi < argc; argi++) {
		if (strcmp(argv[argi], "--help") == 0)
			return 0;
		if (argv[argi][0] != '-')
			break;

		else if (strcmp(argv[argi], "-r") == 0) {
			popts->do_complex = 0;
		}
		else if (strcmp(argv[argi], "-c") == 0) {
			popts->do_complex = 1;
		}
		else if (strcmp(argv[argi], "-circ") == 0) {
			popts->do_circular = 1;
		}

		else {
			return 0;
		}
	}
	if ((argc - argi) != 5)
		return 0;

	if (sscanf(argv[argi], "%f", &popts->num_row_bumps) != 1)
		return 0;
	if (sscanf(argv[argi+1], "%f", &popts->num_col_bumps) != 1)
		return 0;
	if (sscanf(argv[argi+2], "%d", &popts->num_rows) != 1)
		return 0;
	if (sscanf(argv[argi+3], "%d", &popts->num_cols) != 1)
		return 0;
	popts->output_file_name = argv[argi+4];

	return 1;
}
