#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "complex_float.h"
#include "complex_double.h"
#include "tmatrix.h"

typedef struct _opts_t {
	int    do_pack;
	int    num_rows;
	int    num_cols;
	char * input_file_name;
	char * output_file_name;
} opts_t;

static void set_defaults(opts_t * popts);
static int parse_command_line(int argc, char ** argv, opts_t * popts);

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s [-u] {full # rows} {full # cols} {input file} "
		"{output file}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
static void pack_dim_check(
	tmatrix<complex_float> & cpacked,
	tmatrix<complex_float> & cfull)
{
	// E.g. full is 256 x 256, packed is 256 x 129.
	// Here we check 256 == 256.
	if (cfull.get_num_rows() != cpacked.get_num_rows()) {
		fprintf(stderr,
		"Packed and unpacked matrices must have the same number of rows.\n");
		exit(1);
	}

	// E.g. full is 256 x 256, packed is 256 x 129.
	// Here we check 129 == 256 / 2 + 1.
	if (cpacked.get_num_cols() != cfull.get_num_cols()/2 +1) {
		fprintf(stderr,
			"Unrecognized packing format.\n");
		exit(1);
	}
}

// ----------------------------------------------------------------
void svunpack(
	tmatrix<complex_float> & cfull,
	tmatrix<complex_float> & cpacked)
{
	int i, j;
	int half = cfull.get_num_cols() / 2;
	int full_nr = cfull.get_num_rows();
	int full_nc = cfull.get_num_cols();
	int packed_nr = cpacked.get_num_rows();

	pack_dim_check(cpacked, cfull);

	// Let * denote complex conjugation.
	//
	// f00=p00  f01=p01  f02=p02  f03=p03  f04=p04  f05=p03* f06=p02* f07=p01*
	// f10=p10  f11=p11  f12=p12  f13=p13  f14=p14  f15=p73* f16=p72* f17=p71*
	// f20=p20  f21=p21  f22=p22  f23=p23  f24=p24  f25=p63* f26=p62* f27=p61*
	// f30=p30  f31=p31  f32=p32  f33=p33  f34=p34  f35=p53* f36=p52* f37=p51*
	// f40=p40  f41=p41  f42=p42  f43=p43  f44=p44  f45=p43* f46=p42* f47=p41*
	// f50=p50  f51=p51  f52=p52  f53=p53  f54=p54  f55=p33* f56=p32* f57=p31*
	// f60=p60  f61=p61  f62=p62  f63=p63  f64=p64  f65=p23* f66=p22* f67=p21*
	// f70=p70  f71=p71  f72=p72  f73=p73  f74=p74  f75=p13* f76=p12* f77=p11*

	for (i = 0; i < full_nr; i++) {
		for (j = 0; j <= half; j++) {
			cfull[i][j] = cpacked[i][j];
		}
	}

	for (j = half+1; j < full_nc; j++) {
		cfull[0][j].re =  cpacked[0][full_nc-j].re;
		cfull[0][j].im = -cpacked[0][full_nc-j].im;
	}

	for (i = 1; i < packed_nr; i++) {
		for (j = half+1; j < full_nc; j++) {
			cfull[i][j].re =  cpacked[full_nr-i][full_nc-j].re;
			cfull[i][j].im = -cpacked[full_nr-i][full_nc-j].im;
		}
	}
}

// ----------------------------------------------------------------
void svpack(
	tmatrix<complex_float> & cpacked,
	tmatrix<complex_float> & cfull)
{
	int i, j;
	int full_nr = cfull.get_num_rows();
	int full_nc = cfull.get_num_cols();
	int half = full_nc / 2;

	pack_dim_check(cpacked, cfull);

	for (i = 0; i < full_nr; i++) {
		for (j = 0; j <= half; j++) {
			cpacked[i][j] = cfull[i][j];
		}
	}
}

// ----------------------------------------------------------------
int main(
	int     argc,
	char ** argv)
{
	opts_t opts;

	set_defaults(&opts);
	if (!parse_command_line(argc, argv, &opts))
		usage(argv[0]);

	tmatrix<complex_float> cfull  (opts.num_rows, opts.num_cols);
	tmatrix<complex_float> cpacked(opts.num_rows, opts.num_cols/2 + 1);

	if (opts.do_pack) {
		cfull.load_from_binary_file_or_die(opts.input_file_name);
		svpack(cpacked, cfull);
		cpacked.store_to_binary_file_or_die(opts.output_file_name);
	}
	else {
		cpacked.load_from_binary_file_or_die(opts.input_file_name);
		svunpack(cfull, cpacked);
		cfull.store_to_binary_file_or_die(opts.output_file_name);
	}

	return 0;
}

// ----------------------------------------------------------------
static void set_defaults(opts_t * popts)
{
	popts->do_pack          = 1;
	popts->num_rows         = 1;
	popts->num_cols         = 1;
	popts->input_file_name  = 0;
	popts->output_file_name = 0;
}

// ----------------------------------------------------------------
static int parse_command_line(int argc, char ** argv, opts_t * popts)
{
	int argi;
	for (argi = 1; argi < argc; argi++) {
		if (strcmp(argv[argi], "--help") == 0)
			return 0;
		if (argv[argi][0] != '-')
			break;

		else if (strcmp(argv[argi], "-p") == 0)
			popts->do_pack = 1;
		else if (strcmp(argv[argi], "-u") == 0)
			popts->do_pack = 0;

		else {
			return 0;
		}
	}
	if ((argc - argi) != 4)
		return 0;

	if (sscanf(argv[argi], "%d", &popts->num_rows) != 1)
		return 0;
	if (sscanf(argv[argi+1], "%d", &popts->num_cols) != 1)
		return 0;
	popts->input_file_name  = argv[argi+2];
	popts->output_file_name = argv[argi+3];

	return 1;
}
