#include "dda.h"

// ----------------------------------------------------------------
dda_t::dda_t(void)
{
	this->pregs = 0;
	this->nregs = 0;
}

dda_t::dda_t(fixed_point c0)
{
	this->nregs = 1;
	this->pregs = new fixed_point[this->nregs];
	this->pregs[0] = c0;
}

dda_t::dda_t(fixed_point c0, fixed_point c1)
{
	this->nregs = 2;
	this->pregs = new fixed_point[this->nregs];
	this->pregs[0] = c0;
	this->pregs[1] = c1;
}

dda_t::dda_t(fixed_point c0, fixed_point c1, fixed_point c2)
{
	this->nregs = 3;
	this->pregs = new fixed_point[this->nregs];
	this->pregs[0] = c0;
	this->pregs[1] = c1;
	this->pregs[2] = c2;
}

dda_t::dda_t(fixed_point * pseeds, int nseeds)
{
	if (nseeds <= 0) {
		std::cerr << "dda_t:  nseeds must be positive; got "
			<< nseeds << ".\n";
		exit(1);
	}
	this->nregs = nseeds;
	this->pregs = new fixed_point[this->nregs];
	for (int i = 0; i < nseeds; i++)
		this->pregs[i] = pseeds[i];
}

dda_t::dda_t(float c0, int num_bits, int num_frac)
{
	this->nregs = 1;
	this->pregs = new fixed_point[this->nregs];
	this->pregs[0] = fixed_point(c0, num_bits, num_frac);
}

dda_t::dda_t(float c0, float c1, int num_bits, int num_frac)
{
	this->nregs = 2;
	this->pregs = new fixed_point[this->nregs];
	this->pregs[0] = fixed_point(c0, num_bits, num_frac);
	this->pregs[1] = fixed_point(c1, num_bits, num_frac);
}

dda_t::dda_t(float c0, float c1, float c2, int num_bits, int num_frac)
{
	this->nregs = 3;
	this->pregs = new fixed_point[this->nregs];
	this->pregs[0] = fixed_point(c0, num_bits, num_frac);
	this->pregs[1] = fixed_point(c1, num_bits, num_frac);
	this->pregs[2] = fixed_point(c2, num_bits, num_frac);
}

dda_t::dda_t(float * pseeds, int nseeds, int num_bits, int num_frac)
{
	if (nseeds <= 0) {
		std::cerr << "dda_t:  nseeds must be positive; got "
			<< nseeds << ".\n";
		exit(1);
	}
	this->nregs = nseeds;
	this->pregs = new fixed_point[this->nregs];
	for (int i = 0; i < nseeds; i++)
		this->pregs[i] = fixed_point(pseeds[i], num_bits, num_frac);
}

dda_t::dda_t(const dda_t & that)
{
	this->nregs = that.nregs;
	this->pregs = new fixed_point[this->nregs];
	for (int i = 0; i < that.nregs; i++)
		this->pregs[i] = that.pregs[i];
}

dda_t & dda_t::operator=(dda_t that)
{
	if (this->pregs)
		delete [] this->pregs;
	this->nregs = that.nregs;
	this->pregs = new fixed_point[this->nregs];
	for (int i = 0; i < that.nregs; i++)
		this->pregs[i] = that.pregs[i];
	return *this;
}

dda_t::~dda_t(void)
{
	if (this->pregs)
		delete [] this->pregs;
	this->pregs = 0;
	this->nregs = 0;
}

// ----------------------------------------------------------------
void dda_t::set_reg(int regno, fixed_point new_value)
{
	this->bounds_check(regno);
	this->pregs[regno] = new_value;
}

fixed_point dda_t::get_reg(int regno)
{
	this->bounds_check(regno);
	return this->pregs[regno];
}

fixed_point & dda_t::operator[] (int regno)
{
	this->bounds_check(regno);
	return this->pregs[regno];
}

int dda_t::get_num_regs(void)
{
	return this->nregs;
}

// ----------------------------------------------------------------
fixed_point dda_t::get(void)
{
	this->bounds_check(0);
	return this->pregs[0];
}

void dda_t::step(void)
{
	for (int i = 0; i < this->nregs - 1; i++) {
		fixed_point xc0, xc1, xsum;
		xc0 = this->pregs[i];
		xc1 = this->pregs[i+1];
		xsum = xc0 + xc1;
		this->pregs[i] += this->pregs[i+1];
	}
}

fixed_point dda_t::get_and_step(void)
{
	fixed_point rv = this->get();
	this->step();
	return rv;
}

// ----------------------------------------------------------------
std::ostream & operator<<(std::ostream & os, const dda_t & dda)
{
	for (int i = 0; i < dda.nregs; i++) {
		if (i > 0)
			os << " ";
		os << dda.pregs[i];
	}
	return os;
}

// ----------------------------------------------------------------
void dda_t::bounds_check(int regno)
{
	if ((regno < 0) || (regno >= this->nregs)) {
		std::cerr << "dda_t:  register index "
			<< regno << " out of bounds 0:" << this->nregs - 1
			<< ".\n";
		exit(1);
	}
}
