#include <math.h>
#include "fixed_point.h"

#define MASK_FROM_NB(nb) \
	(((nb) == FIXED_POINT_MAX_BITS) ?  ~0LL : ((1LL << (nb)) - 1))

#define mypow(a,b) (exp((b)*log(a)))

// ----------------------------------------------------------------
fixed_point::fixed_point(void)
{
	this->num_bits = 0;
	this->value    = 0;
	this->mask     = 0;
}

fixed_point::fixed_point(int nb)
{
	this->check_num_bits(nb, "fixed_point");

	this->num_bits = nb;
	this->value    = 0;
	this->mask     = 0;
}

fixed_point::fixed_point(float r, int nb, int nf)
{
	this->check_num_bits(nb, "fixed_point");
	this->check_num_frac(nb, nf, "fixed_point");

	this->num_bits = nb;
	this->mask = MASK_FROM_NB(nb);

	if (nf == FIXED_POINT_MAX_BITS)
		this->value = (s64)(r * (1LL << (nf>>1)) * (1LL << (nf>>1)));
	else
		this->value = (s64)(r * (1LL << nf));

	this->value &= this->mask;
}

fixed_point::fixed_point(double r, int nb, int nf)
{
	this->check_num_bits(nb, "fixed_point");
	this->check_num_frac(nb, nf, "fixed_point");

	this->num_bits = nb;
	this->mask = MASK_FROM_NB(nb);

	if (nf == FIXED_POINT_MAX_BITS)
		this->value = (s64)(r * (1LL << (nf>>1)) * (1LL << (nf>>1)));
	else
		this->value = (s64)(r * (1LL << nf));

	this->value &= this->mask;
}

fixed_point::fixed_point(s64 v, int nb)
{
	this->check_num_bits(nb, "fixed_point");
	this->num_bits = nb;
	this->mask = MASK_FROM_NB(nb);
	this->value = v & this->mask;
}

float fixed_point::to_float(int num_frac)
{
	this->check_num_frac(this->num_bits, num_frac, "to_float");
	s64 se = this->sign_extended_s64();
	if (num_frac == FIXED_POINT_MAX_BITS)
		return (float)se / (1LL << (num_frac>>1))
			/ (1LL << (num_frac>>1));
	else
		return (float)se / (1LL << num_frac);
}

double fixed_point::to_double(int num_frac)
{
	this->check_num_frac(this->num_bits, num_frac, "to_double");
	s64 se = this->sign_extended_s64();
	if (num_frac == FIXED_POINT_MAX_BITS)
		return (double)se / (1LL << (num_frac>>1))
			/ (1LL << (num_frac>>1));
	else
		return (double)se / (1LL << num_frac);
}

float fixed_point::to_float(int num_frac, int block_exponent)
{
	this->check_num_frac(this->num_bits, num_frac, "to_float");
	s64 se = this->sign_extended_s64();
	float rv;
	if (num_frac == FIXED_POINT_MAX_BITS)
		rv = (float)se / (1LL << (num_frac>>1))
			/ (1LL << (num_frac>>1));
	else
		rv = (float)se / (1LL << num_frac);
	rv *= mypow(2.0, (float)block_exponent);
	return rv;
}

double fixed_point::to_double(int num_frac, int block_exponent)
{
	this->check_num_frac(this->num_bits, num_frac, "to_double");
	s64 se = this->sign_extended_s64();
	double rv;
	if (num_frac == FIXED_POINT_MAX_BITS)
		rv = (double)se / (1LL << (num_frac>>1))
			/ (1LL << (num_frac>>1));
	else
		rv = (double)se / (1LL << num_frac);
	rv *= mypow(2.0, (double)block_exponent);
	return rv;
}

fixed_point::fixed_point(const fixed_point & that)
{
	this->num_bits = that.num_bits;
	this->value    = that.value;
	this->mask     = that.mask;
}

fixed_point::~fixed_point(void)
{
}

// ----------------------------------------------------------------
void fixed_point::set_num_bits(int nb)
{
	this->check_num_bits(nb, "set_num_bits");
	this->num_bits = nb;
	this->mask = MASK_FROM_NB(nb);
}

void fixed_point::set_value(s64 v)
{
	this->value = v & this->mask;
}

int fixed_point::get_num_bits(void)
{
	return this->num_bits;
}

s64 fixed_point::get_value(void)
{
	return this->value;
}

// ----------------------------------------------------------------
s64 fixed_point::sign_extended_s64(void)
{
	if (this->num_bits == 0)
		return this->value;
	if (this->num_bits == FIXED_POINT_MAX_BITS)
		return this->value;
	s64 sign_bit_mask = 1LL << (this->num_bits - 1);
	if (this->value & sign_bit_mask)
		return this->value | -(1LL << this->num_bits);
	else
		return this->value;
}

// ----------------------------------------------------------------
fixed_point fixed_point::move_num_bits(int nbshift)
{
	fixed_point rv;

	if (nbshift == 0) {
		rv = *this;
	}

	else if (nbshift > 0) {
		if ((this->num_bits + nbshift) > FIXED_POINT_MAX_BITS) {
			std::cerr <<
				"fixed_point::move_num_bits:  overflow.\n";
			exit(1);
		}
		rv.value = this->value << nbshift;
		rv.num_bits = this->num_bits + nbshift;
		rv.mask = MASK_FROM_NB(rv.num_bits);
	}
	else {
		// It is impossible to represent a negative number of bits.
		// We cannot continue.
		if (-nbshift > this->num_bits) {
			std::cerr <<
				"fixed_point::move_num_bits:  underflow.\n";
			exit(1);
		}
		rv.value = this->value >> -nbshift;
		rv.num_bits = this->num_bits + nbshift;
		rv.mask = MASK_FROM_NB(rv.num_bits);
	}

	return rv;
}

// ----------------------------------------------------------------
fixed_point fixed_point::sign_extend_by(int nbdiff)
{
	if (nbdiff < 0) {
		std::cerr << "fixed_point::sign_extend_by:  negative "
			<< "argument disallowed.\n";
		exit(1);
	}
	if (nbdiff == 0) {
		return *this;
	}

	fixed_point rv;
	if ((this->num_bits + nbdiff) > FIXED_POINT_MAX_BITS) {
		std::cerr <<
			"fixed_point::sign_extend_by:  overflow.\n";
		exit(1);
	}
	rv.num_bits = this->num_bits + nbdiff;
	rv.mask = MASK_FROM_NB(rv.num_bits);
	rv.value = this->sign_extended_s64() & rv.mask;
	return rv;
}

// ----------------------------------------------------------------
fixed_point fixed_point::unsigned_extend_by(int nbdiff)
{
	if (nbdiff < 0) {
		std::cerr << "fixed_point::sign_extend_by:  negative "
			<< "argument disallowed.\n";
		exit(1);
	}
	if (nbdiff == 0) {
		return *this;
	}

	fixed_point rv;
	if ((this->num_bits + nbdiff) > FIXED_POINT_MAX_BITS) {
		std::cerr <<
			"fixed_point::sign_extend_by:  overflow.\n";
		exit(1);
	}
	rv.num_bits = this->num_bits + nbdiff;
	rv.mask = MASK_FROM_NB(rv.num_bits);
	rv.value = this->value;
	return rv;
}

// ----------------------------------------------------------------
fixed_point fixed_point::extract_range(int hibit, int lobit)
{
	if (hibit < lobit) {
		std::cerr << "fixed_point::extract_range: "
			<< hibit << " is not > " << lobit << ".\n";
		exit(1);
	}
	this->bounds_check_bit_pos(hibit, "extract_range");
	this->bounds_check_bit_pos(lobit, "extract_range");

	fixed_point rv;
	rv.num_bits = hibit - lobit + 1;
	rv.mask = MASK_FROM_NB(rv.num_bits);
	rv.value = (this->value >> lobit) & rv.mask;

	return rv;
}

s64 fixed_point::extract_range_s64(int hibit, int lobit)
{
	fixed_point fx = this->extract_range(hibit, lobit);
	return fx.value;
}

fixed_point fixed_point::extract_range_with_round(int hibit, int lobit)
{
	if (hibit < lobit) {
		std::cerr << "fixed_point::extract_range: "
			<< hibit << " is not > " << lobit << ".\n";
		exit(1);
	}
	this->bounds_check_bit_pos(hibit, "extract_range_with_round");
	this->bounds_check_bit_pos(lobit, "extract_range_with_round");

	fixed_point rv;
	rv.num_bits = hibit - lobit + 1;
	rv.mask     = MASK_FROM_NB(rv.num_bits);
	rv.value    = this->value;
	// Add 1/2 LSB.  There is a guard bit here since we are doing
	// 64-bit math.
	if (lobit > 0)
		rv.value   += 1LL << (lobit - 1);
	rv.value  >>= lobit;
	rv.value   &= rv.mask;

	return rv;
}

// ----------------------------------------------------------------
fixed_point fixed_point::get_msbs(int nb)
{
	return this->extract_range(this->num_bits - 1, this->num_bits - nb);
}

fixed_point fixed_point::get_lsbs(int nb)
{
	return this->extract_range(nb - 1, 0);
}

s64 fixed_point::get_msbs_s64(int nb)
{
	fixed_point fx = this->get_msbs(nb);
	return fx.value;
}

s64 fixed_point::get_lsbs_s64(int nb)
{
	fixed_point fx = this->get_lsbs(nb);
	return fx.value;
}

// ----------------------------------------------------------------
int fixed_point::get_msb(void)
{
	return this->get_bit(this->num_bits - 1);
}

int fixed_point::get_bit(int which)
{
	this->bounds_check_bit_pos(which, "extract_range");
	return (int)(this->value >> which) & 1;
}

// ----------------------------------------------------------------
fixed_point fixed_point::round(int new_num_bits)
{
	u64 v = this->value;

	if (new_num_bits < this->num_bits) {
		v += (1LL) << (this->num_bits - new_num_bits - 1);
		v >>= this->num_bits - new_num_bits;
	}
	else if (new_num_bits > this->num_bits) {
		v <<= new_num_bits - this->num_bits;
	}

	fixed_point rv(v, new_num_bits);
	return rv;
}

// ----------------------------------------------------------------
fixed_point fixed_point::concatenate(fixed_point that)
{
	this->check_num_bits(this->num_bits + that.num_bits,
		"concatenate");
	fixed_point rv;
	rv.num_bits = this->num_bits + that.num_bits;
	rv.mask = MASK_FROM_NB(rv.num_bits);
	rv.value = (this->value << that.num_bits) | that.value;
	return rv;
}

// ----------------------------------------------------------------
fixed_point & fixed_point::operator=(fixed_point that)
{
	this->num_bits = that.num_bits;
	this->value    = that.value;
	this->mask     = that.mask;
	return *this;
}

fixed_point fixed_point::operator+(fixed_point that)
{
	this->check_dims(that, "operator+");
	s64 a = this->sign_extended_s64();
	s64 b = that.sign_extended_s64();
	s64 c = a + b;
	fixed_point rv;

	rv.value = c;
	rv.num_bits = this->num_bits;
	rv.mask = MASK_FROM_NB(rv.num_bits);
	rv.value &= rv.mask;

	return rv;
}

fixed_point fixed_point::operator-(fixed_point that)
{
	return *this + -that;
}

fixed_point fixed_point::operator*(fixed_point that)
{
	return this->common_mul(that, 1, 1);
}

fixed_point fixed_point::operator-(void)
{
	fixed_point rv(*this);
	rv.value = (-rv.value & rv.mask);
	return rv;
}

fixed_point fixed_point::operator~(void)
{
	fixed_point rv(*this);
	rv.value = (~rv.value & rv.mask);
	return rv;
}

// Pre-increment.
fixed_point fixed_point::operator++(void)
{
	this->value++;
	this->value &= mask;
	return *this;
}

// Post-increment.
fixed_point fixed_point::operator++(int dummy)
{
	fixed_point rv = *this;
	this->value++;
	this->value &= mask;
	return rv;
}

// Pre-decrement.
fixed_point fixed_point::operator--(void)
{
	this->value--;
	this->value &= mask;
	return *this;
}

// Post-decrement.
fixed_point fixed_point::operator--(int dummy)
{
	fixed_point rv = *this;
	this->value--;
	this->value &= mask;
	return rv;
}

fixed_point fixed_point::operator<<(int shamt)
{
	fixed_point rv(*this);
	if (shamt < 0) {
		std::cerr << "fixed_point::operator<<:  negative shifts "
			<< "are disallowed.\n";
		exit(1);
	}
	else if (shamt >= FIXED_POINT_MAX_BITS) {
		// Native operator misbehaves in this case
		rv.value = 0;
	}
	else {
		rv.value <<= shamt;
		rv.value &= rv.mask;
	}
	return rv;
}

fixed_point fixed_point::operator>>(int shamt)
{
	return this->srightshift(shamt);
}

// ----------------------------------------------------------------
fixed_point & fixed_point::operator+=(fixed_point that)
{
	*this = *this + that;
	return *this;
}

fixed_point & fixed_point::operator-=(fixed_point that)
{
	*this = *this - that;
	return *this;
}

fixed_point & fixed_point::operator*=(fixed_point that)
{
	*this = *this * that;
	return *this;
}

fixed_point & fixed_point::operator<<=(int shamt)
{
	*this = *this << shamt;
	return *this;
}

fixed_point & fixed_point::operator>>=(int shamt)
{
	*this = this->srightshift(shamt);
	return *this;
}

// ----------------------------------------------------------------
// Signed right shift
fixed_point fixed_point::srightshift(int shamt)
{
	fixed_point rv(*this);
	if (shamt < 0) {
		std::cerr << "fixed_point::srightshift:  negative shifts "
			<< "are disallowed.\n";
		exit(1);
	}
	else if (shamt >= FIXED_POINT_MAX_BITS) {
		// Native operator misbehaves in this case
		rv.value = 0;
	}
	else {
		rv.value = this->sign_extended_s64();
		rv.value >>= shamt;
		rv.value &= rv.mask;
	}
	return rv;
}

// Unsigned right shift
fixed_point fixed_point::urightshift(int shamt)
{
	fixed_point rv(*this);
	if (shamt < 0) {
		std::cerr << "fixed_point::urightshift:  negative shifts "
			<< "are disallowed.\n";
		exit(1);
	}
	else if (shamt >= FIXED_POINT_MAX_BITS) {
		// Native operator misbehaves in this case
		rv.value = 0;
	}
	else {
		rv.value >>= shamt;
		rv.value &= rv.mask;
	}
	return rv;
}

// ----------------------------------------------------------------
// Equivalent to operator*.
fixed_point fixed_point::ssmul(fixed_point that)
{
	return this->common_mul(that, 1, 1);
}

fixed_point fixed_point::sumul(fixed_point that)
{
	return this->common_mul(that, 1, 0);
}

fixed_point fixed_point::usmul(fixed_point that)
{
	return this->common_mul(that, 0, 1);
}

fixed_point fixed_point::uumul(fixed_point that)
{
	return this->common_mul(that, 0, 0);
}

fixed_point fixed_point::common_mul(fixed_point that, int signed1, int signed2)
{
	fixed_point rv;
	int prod_nb = this->num_bits + that.num_bits;
	this->check_overflow(prod_nb, "multiply");
	s64 a = signed1 ? this->sign_extended_s64() : this->value;
	s64 b = signed2 ?  that.sign_extended_s64() :  that.value;
	rv.value = a * b;
	rv.num_bits = prod_nb;
	rv.mask = MASK_FROM_NB(rv.num_bits);
	rv.value &= rv.mask;
	return rv;
}

// ----------------------------------------------------------------
int fixed_point::operator==(fixed_point that)
{
	fixed_point diff = *this - that;
	if (diff.value == 0LL)
		return 1;
	else
		return 0;
}

int fixed_point::operator!=(fixed_point that)
{
	return !(*this == that);
}

// ----------------------------------------------------------------
std::ostream & operator<<(std::ostream & os, const fixed_point & fx)
{
	os << fx.value;
	return os;
}

std::istream & operator>>(std::istream & is, fixed_point & fx)
{
	is >> fx.value;
	return is;
}

// ----------------------------------------------------------------
int fixed_point::from_string(char * string)
{
	std::istringstream iss(string, std::ios_base::in);
	iss >> *this;
	return iss.fail() ? 0 : 1;
}

// ----------------------------------------------------------------
void fixed_point::check_overflow(int nb, char * function_name)
{
	if ((nb < 0) || (nb > FIXED_POINT_MAX_BITS)) {
		std::cerr << "fixed_point::" << function_name
			<< ": overflow at " << nb << " bits.\n";
		exit(1);
	}
}

// ----------------------------------------------------------------
void fixed_point::check_num_bits(int nb, char * function_name)
{
	if ((nb < 0) || (nb > FIXED_POINT_MAX_BITS)) {
		std::cerr << "fixed_point::" << function_name
			<< ": # bits " << nb << " is not in the range "
			<< "0-" << FIXED_POINT_MAX_BITS << "\n";
		exit(1);
	}
}

// ----------------------------------------------------------------
void fixed_point::check_num_frac(int nb, int nf, char * function_name)
{
	if ((nf < 0) || (nf > FIXED_POINT_MAX_BITS)) {
		std::cerr << "fixed_point::" << function_name
			<< ": # fractional bits " << nb
			<< " is not in the range "
			<< "0-" << FIXED_POINT_MAX_BITS << "\n";
		exit(1);
	}
	if (nf > nb) {
		std::cerr << "fixed_point::" << function_name
			<< ": # fractional bits " << nf
			<< " exceeds total number of bits "
			<< nb << "\n";
		exit(1);
	}
}

// ----------------------------------------------------------------
void fixed_point::check_dims(fixed_point & that, char * function_name)
{
	if (this->num_bits != that.num_bits) {
		std::cerr << "fixed_point::" << function_name
			<< ":  operands must have equal sizes; got "
			<< this->num_bits << ", " << that.num_bits << ".\n";
		exit(1);
	}
}

// ----------------------------------------------------------------
void fixed_point::bounds_check_bit_pos(int which, char * function_name)
{
	if ((which > this->num_bits) || (which < 0)) {
		std::cerr << "fixed_point::" << function_name << ": bit position "
			<< which << " is out of range 0:" << this->num_bits << ".\n";
		exit(1);
	}
}
