#ifndef FIXED_POINT_H
#define FIXED_POINT_H

#include <iostream>
#include <sstream>
#include <iomanip>

typedef long long s64;
typedef unsigned long long u64;

#define FIXED_POINT_MAX_BITS ((int)(8 * sizeof(s64)))

class fixed_point {
public:
	fixed_point(void);
	fixed_point(int num_bits);
	fixed_point(float r, int num_bits, int num_frac);
	fixed_point(double r, int num_bits, int num_frac);
	fixed_point(s64 value, int num_bits);
	float to_float(int num_frac);
	float to_float(int num_frac, int block_exponent);
	double to_double(int num_frac);
	double to_double(int num_frac, int block_exponent);
	fixed_point(const fixed_point & that);
	~fixed_point(void);

	void set_num_bits(int num_bits);
	void set_value(s64 value);

	int get_num_bits(void);
	s64 get_value(void);

	s64 sign_extended_s64(void);

	// Positive argument increases num_bits; negative argument
	// decreases num_bits.  The MSB is moved.  E.g. if current
	// value is 0x10000 with num_bits = 16, then after
	//    c = c.move_num_bits(3);
	// num_bits is 19 and value is 0x80000.
	fixed_point move_num_bits(int nbshift);

	// Negative argument is an error.  Zero argument results in output
	// equal to input.  Positive argument increases num_bits, sign-
	// extending the input.
	fixed_point sign_extend_by(int nbdiff);

	// Negative argument is an error.  Zero argument results in output
	// equal to input.  Positive argument increases num_bits, without
	// sign-extending the input.
	fixed_point unsigned_extend_by(int nbdiff);

	fixed_point extract_range(int hibit, int lobit);
	s64 extract_range_s64(int hibit, int lobit);
	fixed_point extract_range_with_round(int hibit, int lobit);

	fixed_point get_msbs(int num_bits);
	fixed_point get_lsbs(int num_bits);
	s64 get_msbs_s64(int num_bits);
	s64 get_lsbs_s64(int num_bits);

	int get_msb(void);
	int get_bit(int which);

	fixed_point round(int new_num_bits);

	// This is intended to resemble the Verilog concatenation syntax {a, b}.
	// Usage: c = a.concatenate(b).
	fixed_point concatenate(fixed_point that);

	// Addition and subtraction:  Both operands must have the same
	// number of bits.  Please use the above methods to increase/decrease
	// operand size as necessary.

	// Multiplication: Suppose A has m bits and B has n bits.  The product
	// C = A * B has m + n bits.  A signed multiply is done.
	fixed_point & operator=(fixed_point that);
	fixed_point   operator+(fixed_point that);
	fixed_point   operator-(fixed_point that);
	fixed_point   operator*(fixed_point that);
	fixed_point   operator-(void);
	fixed_point   operator~(void);
	fixed_point   operator++(void);      // ++a:  pre-increment
	fixed_point   operator++(int dummy); // a++: post-increment
	fixed_point   operator--(void);      // --a:  pre-decrement
	fixed_point   operator--(int dummy); // a--: post-decrement
	fixed_point   operator<<(int shamt);
	fixed_point   operator>>(int shamt); // Signed right shift

	fixed_point & operator+=(fixed_point that);
	fixed_point & operator-=(fixed_point that);
	fixed_point & operator*=(fixed_point that);
	fixed_point & operator<<=(int shamt);
	fixed_point & operator>>=(int shamt); // Signed right shift

	fixed_point srightshift(int shamt); // Signed right shift
	fixed_point urightshift(int shamt); // Unsigned right shift

	fixed_point ssmul(fixed_point that); // Equivalent to operator*.
	fixed_point sumul(fixed_point that);
	fixed_point usmul(fixed_point that);
	fixed_point uumul(fixed_point that);
	fixed_point common_mul(fixed_point that, int signed1, int signed2);

	int operator==(fixed_point that);
	int operator!=(fixed_point that);

	// For decimal I/O:
	//   std::cout << std::setbase(16) << x;
	// For hex I/O:
	//   std::cout << std::setbase(16) << x;
	friend std::ostream & operator<<(std::ostream & os,
		const fixed_point & fx);

	// num_bits must already be set, or must be set afterwards using
	// set_num_bits().
	friend std::istream & operator>>(std::istream & is,
		fixed_point & fx);
	int from_string(char * string);

private:
	int num_bits;
	s64 value;
	s64 mask;

	void check_overflow(int nb, char * function_name);
	void check_num_bits(int nb, char * function_name);
	void check_num_frac(int nb, int nf, char * function_name);
	void check_dims(fixed_point & that, char * function_name);
	void bounds_check_bit_pos(int which, char * function_name);

};

#endif // FIXED_POINT_H
